// (c) 2006 Philipp Klaus Krause philipp@colecovision.eu

#include "analyzer.h"

#include <limits.h>

#include <map>
#include <set>
#include <string>
#include <fstream>
#include <sstream>

struct huffman_node
{
	class huffman_node *left;
	class huffman_node *right;
	bool end;
	unsigned char value;
	unsigned long id;
};

#define HUFFMAN_LEFT_END 1
#define HUFFMAN_RIGHT_END 2

// Outputs node as C code to stream and destroy tree.
void output(huffman_node *node, std::ostream& stream)
{
	unsigned long right, left;
	unsigned int flags = 0;

	if(node->left->end)
	{
		flags |= HUFFMAN_LEFT_END;
		left = node->left->value;
	}
	else
		left = node->left->id;

	if(node->right->end)
	{
		flags |= HUFFMAN_RIGHT_END;
		right = node->right->value;
	}
	else
		right = node->right->id;

	stream << "/* Node " << node->id << " */ {" << left << ", " << right << "}";

	delete node;
}

int main(int argc, char *argv[])
{
	unsigned long count[UCHAR_MAX + 1];

	// Count character occourances
	analyze_files(argv + 1, count);

	// Create Forest
	std::multimap<unsigned long, huffman_node*> forest;
	{
		unsigned char i;

		for(i = UCHAR_MAX; i >= 1; i--)
		{
			huffman_node *node = new huffman_node;
			node->end = true;
			node->value = i;
			node->id = i;
			forest.insert(std::pair<unsigned long, huffman_node*>(count[i], node));
		}
		huffman_node *node = new huffman_node;
		node->end = true;
		node->value = 0;
		node->id = i;
		forest.insert(std::pair<unsigned long, huffman_node *>(count[0], node));
	}

	// Create tree
	std::set<huffman_node *> inner_nodes;
	std::set<huffman_node *> left_end_nodes;
	std::set<huffman_node *> right_end_nodes;
	std::set<huffman_node *> both_end_nodes;

	while(forest.size() > 1)
	{
		huffman_node *node = new huffman_node;
		std::multimap<unsigned long int, huffman_node*>::iterator i1, i2;
		i1 = i2 = forest.begin();
		i2++;
		node->end = false;
		node->left = i1->second;
		node->right = i2->second;
		forest.erase(i1);
		forest.erase(i2);
		forest.insert(std::pair<unsigned long, huffman_node *>(i1->first + i2->first, node));
		if(node->left->end && node->right->end)
			both_end_nodes.insert(node);
		else if(node->left->end)
			left_end_nodes.insert(node);
		else if(node->right->end)
			right_end_nodes.insert(node);
		else
			inner_nodes.insert(node);
	}

	std::set<huffman_node *>::iterator i;
	std::map<unsigned long, huffman_node *> nodelist;
	unsigned long id = 0;
	unsigned long root, left_end_start, both_end_start, right_end_start;
	
	for(i = inner_nodes.begin();  i != inner_nodes.end(); i++)
		nodelist.insert(std::pair<unsigned long, huffman_node *>(((*i)->id) = id++, *i));
	left_end_start = id;
	for(i = left_end_nodes.begin();  i != left_end_nodes.end(); i++)
		nodelist.insert(std::pair<unsigned long, huffman_node *>(((*i)->id) = id++, *i));
	both_end_start = id;
	for(i = both_end_nodes.begin();  i != both_end_nodes.end(); i++)
		nodelist.insert(std::pair<unsigned long, huffman_node *>(((*i)->id) = id++, *i));
	right_end_start = id;
	for(i = right_end_nodes.begin();  i != right_end_nodes.end(); i++)
		nodelist.insert(std::pair<unsigned long, huffman_node *>(((*i)->id) = id++, *i));
	root = forest.begin()->second->id;

	// Output tree as C code
	std::ofstream codefile("huffman_tree.c", std::ios_base::trunc | std::ios_base::out);
	codefile << "// Huffman tree generated by ColecoVision huffman_analyzer.\n";
	codefile << "// Nodes below HUFFMAN_LS are inner nodes.\n";
	codefile << "// Nodes from HUFFMAN_LS to HUFFMAN_BS - 1 only have a right child node.\n";
	codefile << "// Nodes from HUFFMAN_BS to HUFFMAN_RS - 1 do not have child nodes.\n";
	codefile << "// Nodes from HUFFMAN_RS onward only have a left child node.\n\n";
	codefile << "#include \"cvu_compression.h\"\n\n";
	codefile << "const uint8_t HUFFMAN_ROOT = " << root << ";\n";
	codefile << "const uint8_t HUFFMAN_LS = " << left_end_start << ", HUFFMAN_BS = " << both_end_start << ", HUFFMAN_RS = " << right_end_start << ";\n";
	codefile << "const struct cvu_huffman_node huffman_tree[" << UCHAR_MAX << "] = {\n\n";
	for(std::map<unsigned long int, huffman_node *>::iterator i = nodelist.begin();;)
	{
		output(i->second, codefile);
		i++;
		if(i == nodelist.end())
			break;
		codefile << ",\n";
	}
		
	//output(forest.begin()->second, codefile);
	codefile << "};\n";

	return(0);
}
